{
  "type": "record",
  "Aspect": {
    "name": "incidentExternalLinks"
  },
  "name": "IncidentExternalLinks",
  "namespace": "com.linkedin.pegasus2avro.incident",
  "fields": [
    {
      "type": {
        "type": "array",
        "items": {
          "type": "record",
          "name": "IncidentExternalLink",
          "namespace": "com.linkedin.pegasus2avro.incident",
          "fields": [
            {
              "Searchable": {
                "fieldType": "KEYWORD"
              },
              "type": "string",
              "name": "connectionId",
              "doc": "Unique connection identifier for this integration. Allows linking to multiple instances\nof the same system (e.g., 'primary-jira', 'ops-servicenow', 'critical-pagerduty')."
            },
            {
              "Searchable": {
                "addToFilters": true,
                "fieldName": "externalSystem",
                "fieldType": "KEYWORD",
                "filterNameOverride": "External System"
              },
              "type": "string",
              "name": "system",
              "doc": "The external system name (e.g., 'jira', 'servicenow', 'pagerduty')."
            },
            {
              "Searchable": {
                "fieldName": "externalLinkId",
                "fieldType": "KEYWORD"
              },
              "type": "string",
              "name": "externalId",
              "doc": "The identifier in the external system (e.g., 'PROJ-123' for Jira)."
            },
            {
              "java": {
                "class": "com.linkedin.pegasus2avro.common.url.Url",
                "coercerClass": "com.linkedin.pegasus2avro.common.url.UrlCoercer"
              },
              "type": "string",
              "name": "url",
              "doc": "The direct URL to the external issue/ticket."
            },
            {
              "type": {
                "type": "record",
                "name": "AuditStamp",
                "namespace": "com.linkedin.pegasus2avro.common",
                "fields": [
                  {
                    "type": "long",
                    "name": "time",
                    "doc": "When did the resource/association/sub-resource move into the specific lifecycle stage represented by this AuditEvent."
                  },
                  {
                    "java": {
                      "class": "com.linkedin.pegasus2avro.common.urn.Urn"
                    },
                    "type": "string",
                    "name": "actor",
                    "doc": "The entity (e.g. a member URN) which will be credited for moving the resource/association/sub-resource into the specific lifecycle stage. It is also the one used to authorize the change.",
                    "Urn": "Urn"
                  },
                  {
                    "java": {
                      "class": "com.linkedin.pegasus2avro.common.urn.Urn"
                    },
                    "type": [
                      "null",
                      "string"
                    ],
                    "name": "impersonator",
                    "default": null,
                    "doc": "The entity (e.g. a service URN) which performs the change on behalf of the Actor and must be authorized to act as the Actor.",
                    "Urn": "Urn"
                  },
                  {
                    "type": [
                      "null",
                      "string"
                    ],
                    "name": "message",
                    "default": null,
                    "doc": "Additional context around how DataHub was informed of the particular change. For example: was the change created by an automated process, or manually."
                  }
                ],
                "doc": "Data captured on a resource/association/sub-resource level giving insight into when that resource/association/sub-resource moved into a particular lifecycle stage, and who acted to move it into that specific lifecycle stage."
              },
              "name": "created",
              "doc": "When this link was created."
            },
            {
              "type": "boolean",
              "name": "syncEnabled",
              "default": true,
              "doc": "Whether bidirectional sync is enabled for this link."
            },
            {
              "type": [
                "null",
                "long"
              ],
              "name": "lastSyncTimeMillis",
              "default": null,
              "doc": "Epoch milliseconds of the last successful synchronization."
            }
          ],
          "doc": "A link to an external system (e.g., Jira, ServiceNow, PagerDuty) associated with an incident."
        }
      },
      "name": "links",
      "default": [],
      "doc": "List of links to external systems"
    }
  ],
  "doc": "Links to external systems for this incident (e.g., Jira, ServiceNow, PagerDuty)"
}